package Apache::ReverseLocation;

#---------------------------------------------------------------------
#
# usage: PerlModule Apache::ReverseLocation
#
#---------------------------------------------------------------------

use mod_perl 1.2401;
use Apache::Constants qw(OK DECLINED DECLINE_CMD);
use Apache::ModuleConfig;
use MIME::Base64;
use DynaLoader ();

$Apache::ReverseLocation::VERSION = '0.01';

if ($ENV{MOD_PERL}) {
  no strict;
  @ISA = qw(DynaLoader);
  __PACKAGE__->bootstrap($Apache::ReverseLocation::VERSION);
}

use strict;

# set debug level
#  0 - messages at info or debug log levels
#  1 - verbose output at info or debug log levels
#  2 - really verbose output at info or debug log levels
$Apache::ReverseLocation::DEBUG = 0;

#---------------------------------------------------------------------
# set PerlPostReadRequestHandler globally for now...
#---------------------------------------------------------------------

print STDERR "\tAdding Apache::ReverseLocation::_test\n"
  if $Apache::ReverseLocation::DEBUG;

$Apache::ReadConfig::PerlPostReadRequestHandler = 
  "Apache::ReverseLocation::_test";

sub _test {
#---------------------------------------------------------------------
# test the incoming URI against the reverse regex
# try to add some intelligence so that the user doesn't have to
# dig out the filename 
#---------------------------------------------------------------------

  my $r            = shift;

  my $scfg         = Apache::ModuleConfig->get($r->server);

  my $uri          = $r->uri;

  my $match        = $r->server->server_hostname . $uri;

  foreach my $key (sort keys %{$scfg->{_regex}}) {
    my $regex = $scfg->{_regex}{$key};
    if ($match !~ m!$regex!) {
      # save the real uri
      $r->pnotes(ORIG_URI => $uri);

      # strip all but the filename from the original uri
      my ($file) = $uri =~ m!.*/(.*)!;

      # the new uri is the matched regex key plus any filename
      $r->uri("/${key}/${file}");

      print STDERR "\tURI changed from\n\t\t$uri\n\tto\n\t\t",
                   $r->uri, "\n\tbased on rule:\n\t\t", 
                   $scfg->{_regex}{$key}, "\n"
        if $Apache::ReverseLocation::DEBUG;

      return OK;
    }
  }

  print STDERR "\tURI $uri passed through Apache::ReverseLocation\n"
    if $Apache::ReverseLocation::DEBUG;

  return DECLINED;
}

sub _restore {
#---------------------------------------------------------------------
# restore the original URI
# note that $r->filename is probably messed up, but we try
#---------------------------------------------------------------------

  my $r            = shift;

  $r->uri($r->pnotes("ORIG_URI"));

  print STDERR "\tURI restored to ", $r->pnotes("ORIG_URI"), "\n"
    if $Apache::ReverseLocation::DEBUG;

  return OK;
}

#---------------------------------------------------------------------
# Apache configuration methods
#---------------------------------------------------------------------

sub SERVER_CREATE {
  return bless {}, shift;
}

sub SERVER_MERGE {
  my ($parent, $current) = @_;
  my %new = (%$parent, %$current);

  return bless \%new, ref($parent);
}

sub Location ($$$;*) {
  my ($cfg, $parms, $regex, $fh) = @_;

  return DECLINE_CMD unless $regex =~ m/^!~/;

  my $scfg = Apache::ModuleConfig->get($parms->server);

  print STDERR "\tApache::ReverseLocation intercepted ",
               "<Location $regex\n"
    if $Apache::ReverseLocation::DEBUG;

  _render_location($fh, $regex, $scfg);

}

sub LocationMatch ($$$;*) {
  my ($cfg, $parms, $regex, $fh) = @_;

  return DECLINE_CMD unless $regex =~ m/^!~?/;

  my $scfg = Apache::ModuleConfig->get($parms->server);

  print STDERR "\tApache::ReverseLocation intercepted ",
               "<LocationMatch $regex\n"
    if $Apache::ReverseLocation::DEBUG;

  _render_location($fh, $regex, $parms, $scfg);

}

sub Alias ($$$$;$) {
  my ($cfg, $parms, $not, $regex, $path) = @_;

  return DECLINE_CMD unless $not =~ m/^!~?$/;

  my $scfg = Apache::ModuleConfig->get($parms->server);

  print STDERR "\tApache::ReverseLocation intercepted ",
               "<Alias $not $regex $path>\n"
    if $Apache::ReverseLocation::DEBUG;

  my $host = $parms->server->server_hostname;

  chomp(my $location = $host . encode_base64($regex));

  print STDERR "\twill map URIs to $host matching\n",
               "\t\t\"/$location\"\n\tto path\n",
               "\t\t\"$path\"\n\n";

  @Apache::ReadConfig::Alias = (["/$location" => $path ]);

}

#---------------------------------------------------------------------
# internal methods
#---------------------------------------------------------------------

sub _render_location {

  my ($fh, $regex, $parms, $scfg) = @_;
    
  my %args = (PerlHeaderParserHandler =>
                "Apache::ReverseLocation::_restore");

  while (my $line = <$fh>) {

    last if $line =~ m!</Location(Match)?>!;

    next if $line =~ m!^#!;

    my @args = split / /, $line, 2;
   
    %args = (@args, %args);
   
  }

  $regex =~ s/^!~? *|>$//g;

  my $host = $parms->server->server_hostname;

  chomp(my $location = $host . encode_base64($regex)); 

  $scfg->{_regex}{$location} = $regex;

  if ($Apache::ReverseLocation::DEBUG) {
    print STDERR "\twill redirect URIs to $host not matching\n",
                 "\t\t\"$regex\"\n\tto\n",
                 "\t\t<Location /$location>\n";

    foreach my $key (keys %args) {
      print STDERR "\t\t\t$key $args{$key}\n";
    }

    print STDERR "\t\t<Location>\n\n";
  }

  $Apache::ReadConfig::Location{"/$location"} = \%args;
}

1;

__END__

=head1 NAME

Apache::ReverseLocation - provide for negative URI containers

=head1 SYNOPSIS

httpd.conf:

  PerlModule Apache::ReverseLocation

  <Location !~ "/foo">
    SetHandler perl-script
    PerlHandler My::Handler
  </Location>

  Alias !~ /bar /usr/local/apache/somefiles
  <LocationMatch !~ "/bar">
    SetHandler perl-script
    PerlHandler Another::Handler
  </LocationMatch>

=head1 DESCRIPTION

Apache::ReverseLocation provides negative matching for <Location> and
<LocationMatch> container directives.  They work in the same manner as
the default directives except that the sense of the match is reversed,
similar to Perl's built-in !~ operator, allowing you to set the
behavior of your containers as restrictive instead of permissive.

Apache::ReverseLocation also overrides the mod_alias directive Alias,
which makes it possible to restore the filename (if you need it)
even though the URI is incorrect when translation occurs.  Hopefully
it actually works.

=head1 EXAMPLE

  in httpd.conf:

    PerlModule Apache::ReverseLocation

    <Location !~ "/perl-bin|cgi-bin">
      # this server only allows scripts to be run, nothing else
      PerlAccessHandler Apache::Constants::FORBIDDEN
    </Location>

  if you need to get at $r->filename:
  
  Alias !~ /server-status|/server-info /usr/local/apache/perl-bin
  <Location !~ /server-status|/server-info>
    # if they don't want information, they get a script
    SetHandler perl-script
    PerlHandler Apache::Registry
    Options +ExecCGI
    Order Allow,Deny
    Allow from All
  </Location>

=head1 NOTES

The Alias directive needs to match the regex from the container
directives exactly, otherwise it definitely will not work properly.
It might not work properly anyway...

Verbose debugging is enabled by setting 
$Apache::ReverseLocation::DEBUG=1 in your startup.pl.

The implementation of this module is very incomplete, and as such is
very alpha software.  It has not been tested on multiple platforms
environments for security, stability or other concerns.  It is 
definitely not production quality at this time.

=head1 FEATURES/BUGS

Probably many - it's alphaware.

=head1 SEE ALSO

perl(1), mod_perl(1), Apache(3)

=head1 AUTHOR

Geoffrey Young <geoff@cpan.org>

=head1 COPYRIGHT

Copyright 2001 Geoffrey Young - all rights reserved.

This library is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=cut
